#define AT90S8515
//#define ATmega8515

#define F_CPU            8000000      /* 8Mhz */

#include <avr/io.h>
#include <avr/crc16.h>
#include <avr/eeprom.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
/********  INCLUDE HERE  *************/
#include "bus.c"





/* EEPROM Configuration Address */
#define EE_DEVICE_ADDR    0x00

void init(void);
void PktHandler(void);


int main(void)
{
	unsigned char i, changed;
		
	//start initialization
	init();
	
	//read own address from EEPROM
	eeprom_busy_wait();
	dev_addr = eeprom_read_byte(EE_DEVICE_ADDR);
	
	if (0xFF == dev_addr) {
		eeprom_busy_wait();
		eeprom_write_byte(EE_DEVICE_ADDR, 0x00);
	}
	
	state = 0x00;
	changed = 0x00;
	
	i=0;
	while (i < BUFFER_SIZE) //initialize all buffers to 0x00
	{
		tx_buffer[i]=0x00;
		rx_buffer[i]=0x00;
		rx_input_buffer[i]=0x00;
		i++;
	}
	

	sei();
	
	while(1)
	{
		if (state & RX_PKT_READY) PktHandler();
		
		//the following populates a packet, if changed is set. Of course
		// you can do otherwise, but it is recommended that you check 'state'
		// as below
		if (changed && !(state & (TX_BUF_ACTIVE | TX_PKT_READY)))
		{
			//populate packet, example is given
			//note: CRC16 is calculated automatically
			tx_buffer[PKT_SRC] = dev_addr;
			tx_buffer[PKT_DEST] = 0xFF;
			tx_buffer[PKT_LEN] = 13;			
			tx_buffer[PKT_TYPE] = 'S';
			
			//6 is first data byte
			tx_buffer[6] = 0xFE;
			//BUFFER_SIZE-1 is last data byte
			tx_buffer[BUFFER_SIZE-1] = 0x5C;
			

			changed = 0;
			state = state | TX_PKT_READY;
		}

			
		while (state & TX_PKT_READY)
		{
			if (state & RX_PKT_READY) PktHandler();
			i = xmit_packet();
			if (i) //transmit start failed, wait 10ms and try again
			{
				
				outp(0x00, TCNT1H);
				outp(0x00, TCNT1L);	
				while (TCNT1L < __10_MS)
				{
					if (state & RX_PKT_READY) PktHandler();
				}
				loneliness--;
			}
			else
			{
				cli();
				state = state & (~TX_PKT_READY);
				state = state | TX_BUF_ACTIVE;
				sbi(SER_CONT_REG, UDRIE);
				sei();
			}
		}
	}
}



void PktHandler(void)
{

	unsigned int addr;
	unsigned char i, crc_low, crc_high;
	unsigned short crc;
	crc = 0x0000;
	
	/************** PACKET FILTER *************/

	/* Loopback Test - did we send it? */
	if (rx_buffer[PKT_SRC] == dev_addr) 
	{
		goto	PktExit;
	}
	
	/* Destination Test - is this for us? */
	if ((rx_buffer[PKT_DEST] != 0xFF) && (rx_buffer[PKT_DEST] != dev_addr)) 
	{
		goto	PktExit;
	}
	
	/* CRC16 Test - is the packet intact? */
	for(i=0; i < rx_buffer[PKT_LEN]; i++)
	{
		if ((i != PKT_CRC_H) && (i != PKT_CRC_L))
		{
			crc = _crc16_update(crc, rx_buffer[i]);
		}
	}
	
	crc_low = crc & 0xFF;
	crc_high = crc >> 8;
	
	if((crc_low != rx_buffer[PKT_CRC_L]) || (crc_high != rx_buffer[PKT_CRC_H])) goto PktExit;
	
	/******** END PACKET FILTER *******/
	
	//general purpose ping, answers with lowercase 'a'
	if ('A' == rx_buffer[PKT_TYPE])
	{
		tx_buffer[PKT_SRC] = dev_addr;
		tx_buffer[PKT_DEST] = rx_buffer[PKT_SRC];
		tx_buffer[PKT_LEN] = 0x06;
		tx_buffer[PKT_TYPE] = 'a';
		state = state | TX_PKT_READY;
		goto PktExit;
	} else if ('R' == rx_buffer[PKT_TYPE])
	{
		tx_buffer[PKT_SRC] = dev_addr;
		tx_buffer[PKT_DEST] = rx_buffer[PKT_SRC];
		tx_buffer[PKT_LEN] = 8;
		tx_buffer[PKT_TYPE] = 'r';
		tx_buffer[PKT_CV] = rx_buffer[PKT_CV];
		addr = rx_buffer[PKT_CV] & 0xFF;
		
		tx_buffer[PKT_VALUE] = eeprom_rb(addr);
		state = state | TX_PKT_READY;
		goto PktExit;		
	} else if ('W' == rx_buffer[PKT_TYPE]) {
		
		addr = rx_buffer[PKT_CV] & 0xFF;
		if (!eeprom_is_ready()) eeprom_busy_wait();
		eeprom_wb(addr, rx_buffer[PKT_VALUE]);
		tx_buffer[PKT_VALUE] = rx_buffer[PKT_VALUE];
		
		if (!eeprom_is_ready()) eeprom_busy_wait();
		dev_addr = eeprom_rb(EE_DEVICE_ADDR);
		
		tx_buffer[PKT_SRC] = dev_addr;
		tx_buffer[PKT_DEST] = rx_buffer[PKT_SRC];
		tx_buffer[PKT_LEN] = 0x08;
		tx_buffer[PKT_TYPE] = 'w';
		tx_buffer[PKT_CV] = rx_buffer[PKT_CV];
		
		state = state | TX_PKT_READY;
		goto PktExit;
	}	
	
	PktExit:
	state = state & (~RX_PKT_READY);	
	return;
}


void init(void) //initializes the U(S)ART and the timer
{
	//if ATmega8515 is used, UBRRH must be set 0x00!
	// set baud rate
	UBRR=UART_BAUD_SELECT;
	
	RS485DDR = (1<<RS485TXEN) | (1<<TX); //set TXen pin as output
	RS485PORT = 0x00;	// and low
	
	// Format: 8-N-1
	SER_CONT_REG = ((1<<TXEN) | (1<<RXEN) | (1<<RXCIE));

	// prescaler is 8, timer counts microseconds
	TCCR0 = 0x02;	
	/* prescaler is 1024 for 16bit timer */
	TCCR1A = 0x00;
	TCCR1B = 0x00 | (1<<CS12) | (1<<CS10);
	
		
	ACSR = (1<<ACD);
		
}


